"""Unitsystem compatibility."""

__all__: list[str] = []

from typing import Any, NoReturn

from astropy.coordinates import Angle as AstropyAngle, Distance as AstropyDistance
from astropy.units import Quantity as AstropyQuantity
from plum import conversion_method, dispatch, type_unparametrized as type_up

import quaxed.numpy as jnp
from dataclassish import field_items, replace

from .custom_types import APYUnits
from unxt.dims import dimension_of
from unxt.quantity import AbstractQuantity, AllowValue, BareQuantity, Quantity, ustrip
from unxt.units import unit, unit_of

# ============================================================================
# Value Converter


@dispatch
def convert_to_quantity_value(obj: AstropyQuantity, /) -> NoReturn:
    """Disallow conversion of `AstropyQuantity` to a value.

    >>> import astropy.units as apyu
    >>> from unxt.quantity import convert_to_quantity_value

    >>> try:
    ...     convert_to_quantity_value(apyu.Quantity(1, "m"))
    ... except TypeError as e:
    ...     print(e)
    Cannot convert 'Quantity' to a value.
    For a Quantity, use the `.from_` constructor instead.

    """
    msg = (
        f"Cannot convert {type(obj).__name__!r} to a value. "
        "For a Quantity, use the `.from_` constructor instead."
    )
    raise TypeError(msg)


# ============================================================================
# AbstractQuantity


# -----------------
# Constructor


@AbstractQuantity.from_.dispatch
def from_(
    cls: type[AbstractQuantity], value: AstropyQuantity, /, **kwargs: Any
) -> AbstractQuantity:
    """Construct a `Quantity` from another `Quantity`.

    The `value` is converted to the new `unit`.

    Examples
    --------
    >>> import unxt as u
    >>> import astropy.units as apyu

    >>> u.Quantity.from_(apyu.Quantity(1, "m"))
    Quantity['length'](Array(1., dtype=float32), unit='m')

    """
    u = unit_of(value)
    value = jnp.asarray(ustrip(u, value), **kwargs)
    return cls(value, u)


@AbstractQuantity.from_.dispatch
def from_(
    cls: type[AbstractQuantity], value: AstropyQuantity, u: Any, /, **kwargs: Any
) -> AbstractQuantity:
    """Construct a `Quantity` from another `Quantity`.

    The `value` is converted to the new `unit`.

    Examples
    --------
    >>> import unxt as u
    >>> import astropy.units as apyu

    >>> u.Quantity.from_(apyu.Quantity(1, "m"), "cm")
    Quantity['length'](Array(100., dtype=float32), unit='cm')

    """
    u = unit(u)
    value = jnp.asarray(ustrip(u, value), **kwargs)
    return cls(value, u)


# -----------------
# Conversion Methods


@conversion_method(type_from=AbstractQuantity, type_to=AstropyQuantity)  # type: ignore[arg-type]
def convert_unxt_quantity_to_astropy_quantity(
    q: AbstractQuantity, /
) -> AstropyQuantity:
    """Convert a `unxt.AbstractQuantity` to a `astropy.units.Quantity`.

    Examples
    --------
    >>> import unxt as u
    >>> from astropy.units import Quantity as AstropyQuantity
    >>> from plum import convert

    >>> convert(u.Quantity(1.0, "cm"), AstropyQuantity)
    <Quantity 1. cm>

    """
    u = unit_of(q)
    return AstropyQuantity(ustrip(u, q), u)


@conversion_method(type_from=AbstractQuantity, type_to=AstropyDistance)  # type: ignore[arg-type]
def convert_unxt_quantity_to_astropy_distance(
    q: AbstractQuantity, /
) -> AstropyDistance:
    """Convert a `unxt.AbstractQuantity` to a `astropy.coordinates.Distance`.

    Examples
    --------
    >>> import unxt as u
    >>> from astropy.coordinates import Distance as AstropyDistance
    >>> from plum import convert

    >>> convert(u.Quantity(1.0, "cm"), AstropyDistance)
    <Distance 1. cm>

    """
    u = unit_of(q)
    return AstropyDistance(ustrip(u, q), u)


@conversion_method(type_from=AbstractQuantity, type_to=AstropyAngle)  # type: ignore[arg-type]
def convert_unxt_quantity_to_astropy_angle(q: AbstractQuantity, /) -> AstropyAngle:
    """Convert a `unxt.quantity.AbstractQuantity` to a `astropy.coordinates.Angle`.

    Examples
    --------
    >>> import unxt as u
    >>> from astropy.coordinates import Angle as AstropyAngle
    >>> from plum import convert

    >>> convert(u.Quantity(1.0, "radian"), AstropyAngle)
    <Angle 1. rad>

    """
    u = unit_of(q)
    return AstropyAngle(ustrip(u, q), u)


# ============================================================================
# Quantity


@conversion_method(type_from=AstropyQuantity, type_to=Quantity)  # type: ignore[arg-type]
def convert_astropy_quantity_to_unxt_quantity(q: AstropyQuantity, /) -> Quantity:
    """Convert a `astropy.units.Quantity` to a `unxt.Quantity`.

    Examples
    --------
    >>> from astropy.units import Quantity as AstropyQuantity
    >>> from plum import convert
    >>> from unxt import Quantity

    >>> convert(AstropyQuantity(1.0, "cm"), Quantity)
    Quantity['length'](Array(1., dtype=float32), unit='cm')

    """
    u = unit_of(q)
    return Quantity(ustrip(u, q), u)


# ============================================================================
# BareQuantity


@conversion_method(type_from=AstropyQuantity, type_to=BareQuantity)  # type: ignore[arg-type]
def convert_astropy_quantity_to_unxt_uncheckedquantity(
    q: AstropyQuantity, /
) -> BareQuantity:
    """Convert a `astropy.units.Quantity` to a `unxt.BareQuantity`.

    Examples
    --------
    >>> from astropy.units import Quantity as AstropyQuantity
    >>> from plum import convert
    >>> from unxt.quantity import BareQuantity

    >>> convert(AstropyQuantity(1.0, "cm"), BareQuantity)
    BareQuantity(Array(1., dtype=float32), unit='cm')

    """
    u = unit_of(q)
    return BareQuantity(ustrip(u, q), u)


###############################################################################


@dispatch
def uconvert(u: APYUnits, x: AbstractQuantity, /) -> AbstractQuantity:
    """Convert the quantity to the specified units.

    Examples
    --------
    >>> import astropy.units as apyu
    >>> import unxt as u

    >>> x = u.Quantity(1000, "m")
    >>> u.uconvert(u.unit("km"), x)
    Quantity['length'](Array(1., dtype=float32, ...), unit='km')

    >>> x = u.Quantity([1, 2, 3], "Kelvin")
    >>> with apyu.add_enabled_equivalencies(apyu.temperature()):
    ...     y = x.uconvert("deg_C")
    >>> y
    Quantity['temperature'](Array([-272.15, -271.15, -270.15], dtype=float32, ...), unit='deg_C')

    >>> x = Quantity([1, 2, 3], "radian")
    >>> with apyu.add_enabled_equivalencies(apyu.dimensionless_angles()):
    ...     y = x.uconvert("")
    >>> y
    Quantity['dimensionless'](Array([1., 2., 3.], dtype=float32, ...), unit='')

    """  # noqa: E501
    # Hot-path: if no unit conversion is necessary
    if x.unit == u:
        return x

    # Compute the value. Used in all subsequent branches.
    value = x.unit.to(u, ustrip(x))

    # If the dimensions are the same, we can just replace the value and unit.
    if dimension_of(x.unit) == dimension_of(u):
        return replace(x, value=value, unit=u)

    # If the dimensions are different, we need to create a new quantity since
    # the dimensions can be part of the type. This won't work if the Quantity
    # type itself needs to be changed, e.g. `coordinax.Angle` ->
    # `unxt.Quantity`. These cases are handled separately, in other dispatches.
    fs = dict(field_items(x))
    fs["value"] = value
    fs["unit"] = u

    return type_up(x)(**fs)


@dispatch
def ustrip(u: Any, x: AstropyQuantity) -> Any:
    """Strip the units from the quantity.

    Examples
    --------
    >>> import astropy.units as apyu
    >>> import unxt as u

    >>> x = apyu.Quantity(1000, "m")
    >>> u.ustrip(u.unit("m"), x)
    np.float64(1000.0)

    """
    return x.to_value(u)


@dispatch  # TODO: type annotate by value
def ustrip(flag: type[AllowValue], u: Any, x: AstropyQuantity, /) -> Any:
    """Strip the units from a quantity.

    Examples
    --------
    >>> import unxt as u
    >>> q = u.Quantity(1000, "m")
    >>> u.ustrip(AllowValue, "km", q)
    Array(1., dtype=float32, ...)

    """
    return ustrip(u, x)



"""Functional approach to Quantities."""

__all__ = ["uconvert", "ustrip", "is_unit_convertible"]

from typing import Any

from plum import dispatch


@dispatch.abstract
def uconvert(u: Any, x: Any, /) -> Any:
    """Convert the quantity to the specified units.

    Examples
    --------
    >>> import unxt as u

    >>> q = u.Quantity(1, "km")
    >>> u.uconvert(u.unit("m"), q)
    Quantity['length'](Array(1000., dtype=float32, ...), unit='m')

    >>> u.uconvert("m", q)
    Quantity['length'](Array(1000., dtype=float32, ...), unit='m')

    For further examples, see the other method dispatches.

    """
    raise NotImplementedError  # pragma: no cover


@dispatch.abstract
def ustrip(*args: Any) -> Any:
    """Strip the units from the quantity, first converting if necessary.

    Examples
    --------
    >>> import unxt as u

    >>> q = u.Quantity(1, "km")
    >>> ustrip(u.unit("m"), q)
    Array(1000., dtype=float32, ...)

    >>> u.ustrip("m", q)
    Array(1000., dtype=float32, ...)

    For further examples, see the other method dispatches.

    """
    raise NotImplementedError  # pragma: no cover


@dispatch.abstract
def is_unit_convertible(to_unit: Any, from_: Any, /) -> bool:
    """Check if the units are convertible."""
    return False



"""Benchmark tests for `unxt.unitsystems`."""

import equinox as eqx
import jax
import pytest
from jaxlib.xla_extension import PjitFunction

import unxt as u


@pytest.fixture
def func_unitsystem() -> PjitFunction:
    # The lambda function is necessary because JIT doesn't understand how to
    # introspect the signature of a multiple-dispatch function.
    return eqx.filter_jit(lambda *args: u.unitsystem(*args))


@pytest.fixture
def func_unitsystem_of() -> PjitFunction:
    return eqx.filter_jit(u.unitsystem_of)


@pytest.fixture
def func_equivalent() -> PjitFunction:
    return eqx.filter_jit(u.unitsystems.equivalent)


#####################################################################
# `dimension`

args = [
    (u.unitsystem("kpc", "Myr", "Msun", "radian"),),
    (("kpc", "Myr", "Msun", "radian"),),
    ("kpc", "Myr", "Msun", "radian"),
    (),  # -> dimensionless
    (None,),  # -> dimensionless
    ("galactic",),
    (u.unitsystem("galactic"), "candela"),
    (u.unitsystems.StandardUSysFlag, "galactic"),
    (u.unitsystems.DynamicalSimUSysFlag, "m", "kg"),
]


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="unitsystems", warmup=False, max_time=1.0)
def test_unitsystem(args):
    """Test calling `unxt.unitsystem`."""
    _ = u.unitsystem(*args)


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="unitsystems", warmup=False, max_time=1.0)
def test_unitsystem_jit_compile(func_unitsystem, args):
    """Test the speed of jitting."""
    _ = func_unitsystem.lower(*args).compile()


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="unitsystems", warmup=True, max_time=1.0)
def test_unitsystem_execute(func_unitsystem, args):
    """Test the speed of calling the function."""
    _ = jax.block_until_ready(func_unitsystem(*args))


#####################################################################
# `unitsystem_of`


args = [
    (u.unitsystem("kpc", "Myr", "Msun", "radian"),),
    (2,),  # -> None
]


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="unitsystems", warmup=False, max_time=1.0)
def test_unitsystem_of(args):
    """Test calling `unxt.unitsystem_of`."""
    _ = u.unitsystem_of(*args)


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="unitsystems", warmup=False, max_time=1.0)
def test_unitsystem_of_jit_compile(func_unitsystem_of, args):
    """Test the speed of jitting."""
    _ = func_unitsystem_of.lower(*args).compile()


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="unitsystems", warmup=True, max_time=1.0)
def test_unitsystem_of_execute(func_unitsystem_of, args):
    """Test the speed of calling the function."""
    _ = jax.block_until_ready(func_unitsystem_of(*args))


#####################################################################
# `equivalent`

args = [
    (u.unitsystem("kpc", "Myr", "Msun", "radian"), u.unitsystems.galactic),
]


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="unitsystems", warmup=False, max_time=1.0)
def test_equivalent(args):
    """Test calling `unxt.equivalent`."""
    _ = u.unitsystems.equivalent(*args)


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="unitsystems", warmup=False, max_time=1.0)
def test_equivalent_jit_compile(func_equivalent, args):
    """Test the speed of jitting."""
    _ = func_equivalent.lower(*args).compile()


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="unitsystems", warmup=True, max_time=1.0)
def test_equivalent_execute(func_equivalent, args):
    """Test the speed of calling the function."""
    _ = jax.block_until_ready(func_equivalent(*args))



"""Benchmark tests for `unxt.units`."""

import equinox as eqx
import jax
import pytest
from jaxlib.xla_extension import PjitFunction

import unxt as u


@pytest.fixture
def func_unit() -> PjitFunction:
    return eqx.filter_jit(u.unit)


@pytest.fixture
def func_unit_of() -> PjitFunction:
    # need to filter_jit because arg can be a array or other object
    return eqx.filter_jit(u.unit_of)


#####################################################################
# `unit`

args = [(u.unit("meter"),), ("meter",)]


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="units", warmup=False, max_time=1.0)
def test_unit(args):
    """Test calling `unxt.unit`."""
    _ = u.unit(*args)


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="units", warmup=True, max_time=1.0)
def test_unit_jit_compile(func_unit, args):
    """Test the speed of calling the function."""
    _ = func_unit.lower(*args).compile()


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="units", warmup=True, max_time=1.0)
def test_unit_execute(func_unit, args):
    """Test the speed of calling the function."""
    _ = jax.block_until_ready(func_unit(*args))


#####################################################################
# `unit_of`

args = [(u.unit("meter"),), (u.Quantity(1, "m"),), (2,)]


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="units", warmup=False, max_time=1.0)
def test_unit_of(args):
    """Test calling `unxt.unit_of`."""
    _ = u.unit_of(*args)


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="units", warmup=False, max_time=1.0)
def test_unit_of_jit_compile(func_unit_of, args):
    """Test the speed of jitting a function."""
    _ = func_unit_of.lower(*args).compile()


@pytest.mark.parametrize("args", args, ids=str)
@pytest.mark.benchmark(group="units", warmup=True, max_time=1.0)
def test_unit_of_execute(func_unit_of, args):
    """Test the speed of calling the function."""
    _ = jax.block_until_ready(func_unit_of(*args))



"""`unxt` interoperability with other libraries."""
# ruff:noqa: F401

__all__: list[str] = []

from .optional_deps import OptDeps

# Register interoperability
if OptDeps.ASTROPY.installed:
    from . import unxt_interop_astropy

if OptDeps.GALA.installed:
    from . import unxt_interop_gala

if OptDeps.MATPLOTLIB.installed:
    from . import unxt_interop_mpl as interop_mpl

    interop_mpl.setup_matplotlib_support_for_unxt(enable=True)



r"""Experimental features.

.. warning::

    These features may be removed or changed in the future without notice.

On some occasions JAX's automatic differentiation functions do not work well
with quantities. This is checked by enabling runtime type-checking (see the
docs), which will raise an error if a quantity's units do not match the expected
input / output units of a function. In these cases, you can use the functions in
this module to provide the units to the automatic differentiation functions.
Instead of directly propagating the units through the automatic differentiation
functions, the units are stripped and re-applied, while also being provided
within the function being AD'd.

To import this experimental module

>>> from unxt import experimental

"""
# pylint: disable=import-error

__all__ = ["grad", "hessian", "jacfwd"]

from collections.abc import Callable
from functools import partial
from typing import Any, TypeVar, TypeVarTuple
from typing_extensions import Unpack

import equinox as eqx
import jax
from jaxtyping import ArrayLike
from plum.parametric import type_unparametrized

from .quantity import AbstractQuantity, BareQuantity, ustrip
from .units import AstropyUnits, unit, unit_of

Args = TypeVarTuple("Args")
R = TypeVar("R", bound=AbstractQuantity)


def unit_or_none(obj: Any) -> AstropyUnits | None:
    return obj if obj is None else unit(obj)


def grad(
    fun: Callable[[Unpack[Args]], R],
    argnums: int = 0,
    *,
    units: tuple[AstropyUnits | str, ...],
) -> Callable[[Unpack[Args]], R]:
    """Gradient of a function with units.

    In general, if you can use ``quax.quaxify(jax.grad(func))`` (or the
    syntactic sugar ``quaxed.grad(func)``), that's the better option! The
    difference from those functions is how this units are supported. ``quaxify``
    will directly propagate the units through the automatic differentiation
    functions. But sometimes that doesn't work and we need to strip the units
    and re-apply them. This function does that, using the "units" kwarg.

    See Also
    --------
    jax.grad : The original JAX gradient function.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u

    >>> def cube_volume(x: u.Quantity["length"]) -> u.Quantity["volume"]:
    ...     return x**3

    >>> grad_cube_volume = u.experimental.grad(cube_volume, units=("m",))
    >>> grad_cube_volume(u.Quantity(2.0, "m"))
    Quantity['area'](Array(12., dtype=float32, weak_type=True), unit='m2')

    """
    theunits: tuple[AstropyUnits | None, ...] = tuple(map(unit_or_none, units))

    # Gradient of function, stripping and adding units
    @partial(jax.grad, argnums=argnums)
    def gradfun_mag(*args: Any) -> ArrayLike:
        args_ = (
            (a if unit is None else BareQuantity(a, unit))
            for a, unit in zip(args, theunits, strict=True)
        )
        return ustrip(fun(*args_))  # type: ignore[arg-type]

    def gradfun(*args: Unpack[Args]) -> R:
        # Get the value of the args. They are turned back into Quantity
        # inside the function we are taking the grad of.
        args_ = tuple(  # type: ignore[var-annotated]
            (a if unit is None else ustrip(unit, a))
            for a, unit in zip(args, theunits, strict=True)  # type: ignore[arg-type]
        )
        # Call the grad, returning a Quantity
        value = fun(*args)
        grad_value = gradfun_mag(*args_)
        # Adjust the Quantity by the units of the derivative
        # TODO: get Quantity[unit] / unit2 -> Quantity[unit/unit2] working
        return type_unparametrized(value)(
            grad_value, unit_of(value) / theunits[argnums]
        )

    return gradfun


def jacfwd(
    fun: Callable[[Unpack[Args]], R],
    argnums: int = 0,
    *,
    units: tuple[AstropyUnits | str, ...],
) -> Callable[[Unpack[Args]], R]:
    """Jacobian of ``fun`` evaluated column-by-column using forward-mode AD.

    In general, if you can use `quax.quaxify(jax.jacfwd(func))` (or the
    syntactic sugar `quax.jacfwd(func)`), that's the better option! The
    difference from those functions is how this units are supported. `quaxify`
    will directly propagate the units through the automatic differentiation
    functions. But sometimes that doesn't work and we need to strip the units
    and re-apply them. This function does that, using the ``units`` kwarg.

    See Also
    --------
    jax.jacfwd : The original JAX jacfwd function.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u

    >>> def cubbe_volume(x: u.Quantity["length"]) -> u.Quantity["volume"]:
    ...     return x**3

    >>> jacfwd_cubbe_volume = u.experimental.jacfwd(cubbe_volume, units=("m",))
    >>> jacfwd_cubbe_volume(u.Quantity(2.0, "m"))
    BareQuantity(Array(12., dtype=float32, weak_type=True), unit='m2')

    """
    argnums = eqx.error_if(
        argnums,
        not isinstance(argnums, int),
        "only int argnums are currently supported",
    )

    theunits: tuple[AstropyUnits | None, ...] = tuple(map(unit_or_none, units))

    @partial(jax.jacfwd, argnums=argnums)
    def jacfun_mag(*args: Any) -> R:
        args_ = tuple(
            (a if unit is None else BareQuantity(a, unit))
            for a, unit in zip(args, theunits, strict=True)
        )
        return fun(*args_)  # type: ignore[arg-type]

    def jacfun(*args: Unpack[Args]) -> R:
        # Get the value of the args. They are turned back into Quantity
        # inside the function we are taking the Jacobian of.
        args_ = tuple(  # type: ignore[var-annotated]
            (a if unit is None else ustrip(unit, a))
            for a, unit in zip(args, theunits, strict=True)  # type: ignore[arg-type]
        )
        # Call the Jacobian, returning a Quantity
        value = jacfun_mag(*args_)
        # Adjust the Quantity by the units of the derivative
        # TODO: check the unit correction
        # TODO: get Quantity[unit] / unit2 -> Quantity[unit/unit2] working
        return type_unparametrized(value)(
            ustrip(value), unit_of(value) / theunits[argnums]
        )

    return jacfun


def hessian(
    fun: Callable[[Unpack[Args]], R],
    argnums: int = 0,
    *,
    units: tuple[AstropyUnits | str, ...],
) -> Callable[[Unpack[Args]], R]:
    """Hessian.

    In general, if you can use `quax.quaxify(jax.hessian(func))` (or the
    syntactic sugar `quax.hessian(func)`), that's the better option! The
    difference from those functions is how this units are supported. `quaxify`
    will directly propagate the units through the automatic differentiation
    functions. But sometimes that doesn't work and we need to strip the units
    and re-apply them. This function does that, using the ``units`` kwarg.

    See Also
    --------
    jax.hessian : The original JAX hessian function.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u

    >>> def cubbe_volume(x: u.Quantity["length"]) -> u.Quantity["volume"]:
    ...     return x**3

    >>> hessian_cubbe_volume = u.experimental.hessian(cubbe_volume, units=("m",))
    >>> hessian_cubbe_volume(u.Quantity(2.0, "m"))
    BareQuantity(Array(12., dtype=float32, weak_type=True), unit='m')

    """
    theunits: tuple[AstropyUnits, ...] = tuple(map(unit_or_none, units))

    @partial(jax.hessian)
    def hessfun_mag(*args: Any) -> R:
        args_ = tuple(
            (a if unit is None else BareQuantity(a, unit))
            for a, unit in zip(args, theunits, strict=True)
        )
        return fun(*args_)  # type: ignore[arg-type]

    def hessfun(*args: Unpack[Args]) -> R:
        # Get the value of the args. They are turned back into Quantity
        # inside the function we are taking the hessian of.
        args_ = tuple(  # type: ignore[var-annotated]
            (a if unit is None else ustrip(unit, a))
            for a, unit in zip(args, units, strict=True)  # type: ignore[arg-type]
        )
        # Call the hessian, returning a Quantity
        value = hessfun_mag(*args_)
        # Adjust the Quantity by the units of the derivative
        # TODO: check the unit correction
        # TODO: get Quantity[unit] / unit2 -> Quantity[unit/unit2] working
        return type_unparametrized(value)(
            ustrip(value), unit_of(value) / theunits[argnums] ** 2
        )

    return hessfun



"""Functional approach to Quantities."""

__all__: list[str] = []


from typing import Any

from astropy.units import UnitConversionError
from jaxtyping import Array
from plum import dispatch

from .api import ustrip
from .base import AbstractQuantity
from unxt._src.units import AstropyUnits
from unxt.dims import AbstractDimension, dimension_of
from unxt.units import unit
from unxt.unitsystems import AbstractUnitSystem

# ===================================================================
# Get dimensions


@dispatch
def dimension_of(obj: AbstractQuantity, /) -> AbstractDimension:
    """Return the dimension of a quantity.

    Examples
    --------
    >>> from unxt import dimension_of, Quantity
    >>> q = Quantity(1, "m")
    >>> dimension_of(q)
    PhysicalType('length')

    """
    return dimension_of(obj.unit)


# ===================================================================
# Get units


@dispatch
def unit_of(obj: AbstractQuantity, /) -> AstropyUnits:
    """Return the units of an object.

    Examples
    --------
    >>> from unxt import unit_of, Quantity
    >>> q = Quantity(1, "m")
    >>> unit_of(q)
    Unit("m")

    """
    return obj.unit


# ===================================================================
# Convert units


@dispatch
def uconvert(ustr: str, x: AbstractQuantity, /) -> AbstractQuantity:
    """Convert the quantity to the specified units.

    Examples
    --------
    >>> from unxt import Quantity, units

    >>> x = Quantity(1000, "m")
    >>> uconvert("km", x)
    Quantity['length'](Array(1., dtype=float32, ...), unit='km')

    """
    return uconvert(unit(ustr), x)


@dispatch
def uconvert(usys: AbstractUnitSystem, x: AbstractQuantity, /) -> AbstractQuantity:
    """Strip the units from the quantity.

    Examples
    --------
    >>> from unxt import Quantity, units
    >>> from unxt.unitsystems import galactic

    >>> q = Quantity(1e17, "km")
    >>> uconvert(galactic, q)
    Quantity['length'](Array(3.2407792, dtype=float32, ...), unit='kpc')

    """
    return uconvert(usys[dimension_of(x)], x)


# ===================================================================
# Strip units


@dispatch
def ustrip(x: AbstractQuantity, /) -> Array:
    """Strip the units from the quantity.

    Examples
    --------
    >>> import unxt as u

    >>> q = u.Quantity(1000, "m")
    >>> u.ustrip(q)
    Array(1000, dtype=int32, weak_type=True)

    >>> u.ustrip(q) is q.value
    True

    """
    return x.value


@dispatch
def ustrip(u: AstropyUnits, x: AbstractQuantity, /) -> Array:
    """Strip the units from the quantity.

    Examples
    --------
    >>> import unxt as u

    >>> q = u.Quantity(1000, "m")
    >>> u.ustrip(u.unit("km"), q)
    Array(1., dtype=float32, ...)

    """
    return ustrip(uconvert(u, x))


@dispatch
def ustrip(u: str, x: AbstractQuantity, /) -> Array:
    """Strip the units from the quantity.

    Examples
    --------
    >>> import unxt as u

    >>> q = u.Quantity(1000, "m")
    >>> u.ustrip("km", q)
    Array(1., dtype=float32, ...)

    """
    return ustrip(unit(u), x)


@dispatch
def ustrip(u: AbstractUnitSystem, x: AbstractQuantity, /) -> Array:
    """Strip the units from the quantity.

    Examples
    --------
    >>> import unxt as u
    >>> from unxt.unitsystems import galactic

    >>> q = u.Quantity(1e17, "km")
    >>> u.ustrip(galactic, q)
    Array(3.2407792, dtype=float32, weak_type=True)

    """
    return ustrip(u[dimension_of(x)], x)


# ============================================================================


@dispatch
def is_unit_convertible(to_unit: Any, from_: Any, /) -> bool:
    """Check if a unit can be converted to another unit.

    Parameters
    ----------
    to_unit : Any
        The unit to convert to. Converted to a unit object using `unxt.unit`.
    from_ : Any
        The unit to convert from. Converted to a unit object using `unxt.unit`,
        Note this means it also support `Quantity` objects and many others.

    Examples
    --------
    >>> from unxt import is_unit_convertible
    >>> is_unit_convertible("cm", "m")
    True

    >>> is_unit_convertible("m", "Gyr")
    False

    """
    to_u = unit(to_unit)
    from_u = unit(from_)
    try:
        from_u.to(to_u)
    except UnitConversionError:
        return False
    return True


@dispatch
def is_unit_convertible(to_unit: Any, from_: AbstractQuantity, /) -> bool:
    """Check if a Quantity can be converted to another unit.

    Examples
    --------
    >>> from unxt import Quantity, is_unit_convertible
    >>> q = Quantity(1, "m")

    >>> is_unit_convertible("cm", q)
    True

    >>> is_unit_convertible("Gyr", q)
    False

    """
    return is_unit_convertible(to_unit, from_.unit)



"""Systems of units.

A unit system is a collection of units that are used together. In a unit system
there are base units and derived units. Base units are the fundamental units of
the system and derived units are constructed from the base units. For example,
in the SI system, the base units are the meter, kilogram, second, ampere,
kelvin, mole, and candela. Derived units are constructed from these base units,
for example, the newton is a derived unit of force.

`unxt` provides powerful tools for defining and working with unit systems. Unit
systems can be statically defined (useful for many tools and development
environments) or dynamically defined (useful for interactive environments and
Python's general dynamism). Unit systems can be extended, compared, and used for
decomposing units on quantities. There are many more features and tools for
working with unit systems in `unxt`.

"""
# pylint: disable=duplicate-code

__all__ = [
    "unitsystem",
    "unitsystem_of",
    # classes
    "AbstractUnitSystem",
    "UNITSYSTEMS_REGISTRY",
    "DimensionlessUnitSystem",
    "LTMAUnitSystem",
    # unit system instance
    "galactic",
    "dimensionless",
    "solarsystem",
    "si",
    "cgs",
    # unit system alias
    "NAMED_UNIT_SYSTEMS",
    # unit system flags
    "AbstractUSysFlag",
    "StandardUSysFlag",
    "DynamicalSimUSysFlag",
    # functions
    "equivalent",
]

from jaxtyping import install_import_hook

from .setup_package import RUNTIME_TYPECHECKER

with install_import_hook("unxt.unitsystems", RUNTIME_TYPECHECKER):
    from ._src.unitsystems import (
        NAMED_UNIT_SYSTEMS,
        UNITSYSTEMS_REGISTRY,
        AbstractUnitSystem,
        AbstractUSysFlag,
        DimensionlessUnitSystem,
        DynamicalSimUSysFlag,
        LTMAUnitSystem,
        StandardUSysFlag,
        cgs,
        dimensionless,
        equivalent,
        galactic,
        si,
        solarsystem,
        unitsystem,
        unitsystem_of,
    )

# Clean up the namespace
del install_import_hook, RUNTIME_TYPECHECKER



"""Test that `UncheckedQuantity` functions but is deprecated."""

import pytest

import quaxed.numpy as jnp

from unxt.quantity import BareQuantity, UncheckedQuantity


def test_unchecked_quantity_deprecated():
    with pytest.warns(DeprecationWarning):
        UncheckedQuantity(1, "m")


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_uncheckedquantity_works_like_barequantity():
    # Test creation
    uq1 = UncheckedQuantity(1, "m")
    bq1 = BareQuantity(1, "m")
    assert uq1.value == bq1.value
    assert uq1.unit == bq1.unit

    # Test addition
    uq2 = UncheckedQuantity(2, "m")
    bq2 = BareQuantity(2, "m")
    addu = uq1 + uq2
    addb = bq1 + bq2
    assert addu.value == addb.value
    assert addu.unit == addb.unit

    # Test multiplication
    mulu = uq1 * 2
    mulb = bq1 * 2
    assert mulu.value == mulb.value
    assert mulu.unit == mulb.unit

    # Test numpy operations
    addu = jnp.add(uq1, uq2)
    addb = jnp.add(bq1, bq2)
    assert jnp.array_equal(addu.value, addb.value)
    assert addu.unit == addb.unit



"""Interoperability with `matplotlib`."""
# pylint: disable=import-error

# This module was adapted from both the `astropy` and `pint` package's
# implementations.

__all__ = ["UnxtConverter", "setup_matplotlib_support_for_unxt"]


from collections.abc import Iterable, Sized
from dataclasses import dataclass, field
from typing import Any

import matplotlib.units
from jaxtyping import Array
from matplotlib.axes import Axes

from zeroth import zeroth

from unxt.quantity import AbstractQuantity, BareQuantity, ustrip


@dataclass
class UnxtConverter(matplotlib.units.ConversionInterface):  # type: ignore[misc]
    """Support `unxt` in `matplotlib`'s unit conversion framework.

    This class is a subclass of `matplotlib.units.ConversionInterface`
    and is used to convert `unxt.Quantity` instances for use with
    `matplotlib`.

    """

    axisinfo_kw: dict[str, Any] = field(
        default_factory=lambda: {"format": "latex_inline"}
    )
    """Keyword arguments to use when making the :meth:`matplotlib.units.AxisInfo`."""

    def convert(self, obj: Any, unit: Any, axis: Axes) -> Array | list[Array]:
        """Convert *obj* using *unit* for the specified *axis*."""
        # Hot-path Quantity
        if isinstance(obj, AbstractQuantity):
            return ustrip(unit, obj)
        # Need to recurse (singly) into iterables
        if isinstance(obj, Iterable):
            return [self._convert_value(v, unit, axis) for v in obj]

        return self._convert_value(obj, unit, axis)

    @staticmethod
    def _convert_value(obj: Any, unit: Any, axis: Axes) -> Array:
        """Handle converting using attached unit or falling back to axis units."""
        if isinstance(obj, AbstractQuantity):
            return ustrip(unit, obj)

        return BareQuantity.from_(obj, axis.get_units()).unstrip(unit)

    def axisinfo(self, unit: Any, _: Axes) -> matplotlib.units.AxisInfo:
        """Return axis information for this particular unit."""
        fmt = self.axisinfo_kw.get("format", "latex_inline")
        return matplotlib.units.AxisInfo(label=unit.to_string(fmt))

    @staticmethod
    def default_units(x: Any, _: Axes) -> Any:
        """Get the default unit to use for the given combination of unit and axis."""
        if hasattr(x, "unit"):
            return x.unit
        if isinstance(x, Iterable) and isinstance(x, Sized):
            x = zeroth(x)
        return getattr(x, "units", None)


def setup_matplotlib_support_for_unxt(*, enable: bool = True) -> None:
    """Set up matplotlib's unit support for `unxt`.

    :param enable: Whether support should be enabled or disabled.

    """
    if enable:
        matplotlib.units.registry[AbstractQuantity] = UnxtConverter()
    else:
        matplotlib.units.registry.pop(AbstractQuantity, None)



"""Units objects in JAX.

Copyright (c) 2023 Galactic Dynamics. All rights reserved.
"""

__all__ = ["dimension", "dimension_of"]

from typing import Any, TypeAlias

import astropy.units as apyu
from plum import dispatch

AbstractDimension: TypeAlias = apyu.PhysicalType


@dispatch.abstract
def dimension(obj: Any, /) -> AbstractDimension:
    """Construct the dimension.

    .. note::

        This function uses multiple dispatch. Dispatches made in other modules
        may not be included in the rendered docs. To see the full range of
        options, execute ``unxt.dims.dimension.methods`` in an interactive
        Python session.

    """


@dispatch.abstract
def dimension_of(obj: Any, /) -> AbstractDimension:
    """Return the dimension of the given units.

    .. note::

        This function uses multiple dispatch. Dispatches made in other modules
        may not be included in the rendered docs. To see the full range of
        options, execute ``unxt.dimension_of.methods`` in an interactive Python
        session.

    """



"""Built-in dimensions."""

__all__: list[str] = []

from unxt._src.dimensions import dimension

dimensionless = dimension("dimensionless")
length = dimension("length")
mass = dimension("mass")
time = dimension("time")
speed = dimension("speed")
angle = dimension("angle")
amount = dimension("amount of substance")
current = dimension("electrical current")
temperature = dimension("temperature")
luminous_intensity = dimension("luminous intensity")
pressure = dimension("pressure")
force = dimension("force")
energy = dimension("energy")
dynamic_viscosity = dimension("dynamic viscosity")
kinematic_viscosity = dimension("kinematic viscosity")



"""Dimensions module.

This is the private implementation of the dimensions module.

"""

__all__ = ["AbstractDimension", "dimension", "dimension_of"]

from .api import AbstractDimension, dimension, dimension_of

# Register the dispatches
# isort: split
from . import core  # noqa: F401



"""Realizations of unit systems."""
# pylint: disable=no-member

__all__ = [
    # unit system instance
    "galactic",
    "dimensionless",
    "solarsystem",
    "si",
    "cgs",
    # unit system alias
    "NAMED_UNIT_SYSTEMS",
]

from .base import AbstractUnitSystem
from .builtin import (
    CGSUnitSystem,
    DimensionlessUnitSystem,
    LTMAUnitSystem,
    SIUnitSystem,
)
from unxt.units import unit

# Dimensionless. This is a singleton.
dimensionless = DimensionlessUnitSystem()

# Galactic unit system
galactic = LTMAUnitSystem(
    length=unit("kpc"),
    time=unit("Myr"),
    mass=unit("Msun"),
    angle=unit("radian"),
)

# Solar system units
solarsystem = LTMAUnitSystem(
    length=unit("au"), time=unit("yr"), mass=unit("Msun"), angle=unit("radian")
)

#: International System of Units
si = SIUnitSystem(
    length=unit("meter"),
    time=unit("second"),
    mass=unit("kilogram"),
    electric_current=unit("ampere"),
    temperature=unit("Kelvin"),
    amount=unit("mole"),
    luminous_intensity=unit("candela"),
    angle=unit("radian"),
)

#: Centimeter, gram, second unit system
cgs = CGSUnitSystem(
    length=unit("centimeter"),
    time=unit("second"),
    mass=unit("gram"),
    angle=unit("radian"),
    force=unit("dyne"),
    energy=unit("erg"),
    pressure=unit("barye"),
    dynamic_viscosity=unit("poise"),
    kinematic_viscosity=unit("stokes"),
)


#: Named unit systems
NAMED_UNIT_SYSTEMS: dict[str, AbstractUnitSystem] = {
    "galactic": galactic,
    "solarsystem": solarsystem,
    "dimensionless": dimensionless,
    "si": si,
    "cgs": cgs,
}



"""Test the Array API."""
# pylint: disable=import-error, too-many-lines
# ruff:noqa: E402

import re

import pytest
from jax import Array
from jax._src.numpy.setops import (
    _UniqueAllResult,
    _UniqueCountsResult,
    _UniqueInverseResult,
)
from jax.numpy import iinfo as IInfo  # noqa: N812
from numpy import finfo as FInfo  # noqa: N812

import quaxed.numpy as jnp

import unxt as u

# =============================================================================
# Constants


def test_e():
    """Test `e`."""
    assert not isinstance(jnp.e, u.Quantity)


def test_inf():
    """Test `inf`."""
    assert not isinstance(jnp.inf, u.Quantity)


def test_nan():
    """Test `nan`."""
    assert not isinstance(jnp.nan, u.Quantity)


def test_newaxis():
    """Test `newaxis`."""
    assert not isinstance(jnp.newaxis, u.Quantity)


def test_pi():
    """Test `pi`."""
    assert not isinstance(jnp.pi, u.Quantity)


# =============================================================================
# Creation functions

# -----------------------------------------------
# arange


def test_arange_start():
    """Test `arange`."""
    # -----------------------
    got = jnp.arange(u.Quantity(10, "m"))
    expected = u.Quantity(jnp.arange(10), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)

    # -----------------------
    got = jnp.arange(start=u.Quantity(10, "m"))

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_arange_stop():
    """Test `arange`."""
    start = u.Quantity(2, "m")
    stop = u.Quantity(10, "km")
    got = jnp.arange(start, stop)
    expected = u.Quantity(jnp.arange(start.value, u.ustrip(start.unit, stop)), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_arange_step():
    """Test `arange`."""
    start = u.Quantity(2, "m")
    stop = u.Quantity(10, "km")
    step = u.Quantity(2, "m")
    got = jnp.arange(start, stop, step)
    expected = u.Quantity(
        jnp.arange(start.value, u.ustrip(start.unit, stop), u.ustrip(start.unit, step)),
        "m",
    )

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# -----------------------------------------------


def test_asarray():
    """Test `asarray`."""
    # TODO: test the dtype, device, copy arguments
    x = [1, 2, 3]
    got = jnp.asarray(u.Quantity(x, "m"))
    expected = u.Quantity(jnp.asarray(x), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# -----------------------------------------------


@pytest.mark.xfail(reason="returns a jax.Array")
def test_empty():
    """Test `empty`."""
    # TODO: test the dtype, device arguments
    got = jnp.empty((2, 3))
    assert isinstance(got, u.Quantity)


# -----------------------------------------------


def test_empty_like():
    """Test `empty_like`."""
    x = jnp.asarray([1, 2, 3], dtype=float)
    q = u.Quantity(x, "m")
    got = jnp.empty_like(q)
    expected = u.Quantity(jnp.empty_like(x), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# -----------------------------------------------


@pytest.mark.xfail(reason="returns a jax.Array")
def test_eye():
    """Test `eye`."""
    got = jnp.eye(3)

    assert isinstance(got, u.Quantity)


# -----------------------------------------------


@pytest.mark.skip("TODO")
def test_from_dlpack():
    """Test `from_dlpack`."""
    raise NotImplementedError


# -----------------------------------------------


@pytest.mark.xfail(reason="returns a jax.Array")
def test_full():
    """Test `full`."""
    got = jnp.full((2, 3), 1.0)

    assert isinstance(got, u.Quantity)


# -----------------------------------------------


def test_full_like_single_arg():
    """Test `full_like`."""
    x = jnp.asarray([1, 2, 3], dtype=float)
    q = u.Quantity(x, "m")
    got = jnp.full_like(q, fill_value=1.0)
    expected = u.Quantity(jnp.full_like(x, fill_value=1.0), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_full_like_double_arg():
    """Test `full_like`."""
    x = jnp.asarray([1, 2, 3], dtype=float)
    q = u.Quantity(x, "m")
    got = jnp.full_like(q, 1.0)
    expected = u.Quantity(jnp.full_like(x, 1.0), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_full_like_dtype():
    """Test `full_like`."""
    x = jnp.asarray([1, 2, 3], dtype=float)
    q = u.Quantity(x, "m")
    got = jnp.full_like(q, 1.0, dtype=int)
    expected = u.Quantity(jnp.full_like(q.value, 1.0, dtype=int), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)
    assert got.value.dtype == expected.value.dtype


# -----------------------------------------------


def test_linspace():
    """Test `linspace`."""
    # TODO: test the dtype, device, endpoint arguments
    got = jnp.linspace(u.Quantity(0.0, "m"), u.Quantity(10.0, "m"), 11)
    expected = u.Quantity(jnp.linspace(0.0, 10.0, 11), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# -----------------------------------------------


def test_meshgrid():
    """Test `meshgrid`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")

    got1, got2 = jnp.meshgrid(x, y)
    exp1, exp2 = jnp.meshgrid(x.value, y.value)

    assert isinstance(got1, u.Quantity)
    assert jnp.array_equal(got1.value, exp1)

    assert isinstance(got2, u.Quantity)
    assert jnp.array_equal(got2.value, exp2)


# -----------------------------------------------


@pytest.mark.xfail(reason="returns a jax.Array")
def test_ones():
    """Test `ones`."""
    assert isinstance(jnp.ones((2, 3)), u.Quantity)


# -----------------------------------------------


def test_ones_like():
    """Test `ones_like`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.ones_like(x)
    expected = u.Quantity(jnp.ones_like(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# -----------------------------------------------


def test_tril():
    """Test `tril`."""
    x = u.Quantity([[1, 2, 3], [4, 5, 6], [7, 8, 9]], "m")
    got = jnp.tril(x)
    expected = u.Quantity(jnp.tril(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# -----------------------------------------------


def test_triu():
    """Test `triu`."""
    x = u.Quantity([[1, 2, 3], [4, 5, 6], [7, 8, 9]], "m")
    got = jnp.triu(x)
    expected = u.Quantity(jnp.triu(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# -----------------------------------------------


@pytest.mark.xfail(reason="returns a jax.Array")
def test_zeros():
    """Test `zeros`."""
    assert isinstance(jnp.zeros((2, 3)), u.Quantity)


# -----------------------------------------------


def test_zeros_like():
    """Test `zeros_like`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.zeros_like(x)
    expected = u.Quantity(jnp.zeros_like(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# =============================================================================
# Data-type functions


def test_astype():
    """Test `astype`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.astype(x, jnp.float32)
    expected = u.Quantity(jnp.asarray(x.value, dtype=jnp.float32), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.skip("TODO")
def test_can_cast():
    """Test `can_cast`."""


def test_finfo():
    """Test `finfo`."""
    got = jnp.finfo(jnp.float32)
    expected = jnp.finfo(jnp.float32)

    assert isinstance(got, FInfo)
    for attr in ("bits", "eps", "max", "min", "smallest_normal", "dtype"):
        assert getattr(got, attr) == getattr(expected, attr)


def test_iinfo():
    """Test `iinfo`."""
    got = jnp.iinfo(jnp.int32)
    expected = jnp.iinfo(jnp.int32)

    assert isinstance(got, IInfo)
    for attr in ("kind", "bits", "min", "max", "dtype"):
        assert getattr(got, attr) == getattr(expected, attr)


def test_isdtype():
    """Test `isdtype`."""
    # True by definition


def test_result_type():
    """Test `result_type`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.result_type(x, y)
    expected = jnp.result_type(x.value, y.value)

    assert isinstance(got, jnp.dtype)
    assert got == expected


# =============================================================================
# Elementwise functions


def test_abs():
    """Test `abs`."""
    x = u.Quantity([-1, 2, -3], "m")
    got = jnp.abs(x)
    expected = u.Quantity(jnp.abs(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_acos():
    """Test `acos`."""
    x = u.Quantity(jnp.asarray([-1, 0, 1], dtype=float), "")
    got = jnp.acos(x)
    expected = u.Quantity(jnp.acos(x.value), "rad")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_acosh():
    """Test `acosh`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "")
    got = jnp.acosh(x)
    expected = u.Quantity(jnp.acosh(x.value), "rad")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_add():
    """Test `add`."""
    # Adding two quantities
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.add(x, y)
    expected = u.Quantity(jnp.add(x.value, y.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)

    # Adding a quantity and non-quantity
    match = re.escape("'m' (length) and '' (dimensionless) are not convertible")
    with pytest.raises(ValueError, match=match):
        jnp.add(x.value, y)

    with pytest.raises(ValueError, match=match):
        jnp.add(x, y.value)

    # Add a non-quantity and dimensionless quantity
    got = jnp.add(x.value, u.Quantity(1.0, ""))
    expected = u.Quantity(x.value + 1, "")
    assert jnp.array_equal(got.value, expected.value)

    got = jnp.add(u.Quantity(1.0, ""), y.value)
    expected = u.Quantity(1 + y.value, "")
    assert jnp.array_equal(got.value, expected.value)


def test_asin():
    """Test `asin`."""
    x = u.Quantity(jnp.asarray([-1, 0, 1], dtype=float), "")
    got = jnp.asin(x)
    expected = u.Quantity(jnp.asin(x.value), "rad")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_asinh():
    """Test `asinh`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "")
    got = jnp.asinh(x)
    expected = u.Quantity(jnp.asinh(x.value), "rad")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_atan():
    """Test `atan`."""
    x = u.Quantity(jnp.asarray([-1, 0, 1], dtype=float), "")
    got = jnp.atan(x)
    expected = u.Quantity(jnp.atan(x.value), "rad")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_atan2():
    """Test `atan2`."""
    x = u.Quantity(jnp.asarray([-1, 0, 1], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.atan2(x, y)
    expected = u.Quantity(jnp.atan2(x.value, y.value), "rad")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_atanh():
    """Test `atanh`."""
    x = u.Quantity(jnp.asarray([-1, 0, 1], dtype=float), "")
    got = jnp.atanh(x)
    expected = u.Quantity(jnp.atanh(x.value), "rad")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_bitwise_and():
    """Test `bitwise_and`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=int), "")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=int), "")
    got = jnp.bitwise_and(x, y)
    expected = jnp.bitwise_and(x.value, y.value)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, expected)


@pytest.mark.xfail(reason="TODO")
def test_bitwise_left_shift():
    """Test `bitwise_left_shift`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=int), "")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=int), "")
    got = jnp.bitwise_left_shift(x, y)
    expected = u.Quantity(jnp.bitwise_left_shift(x.value, y.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_bitwise_invert():
    """Test `bitwise_invert`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=int), "")
    got = jnp.bitwise_invert(x)
    expected = u.Quantity(jnp.bitwise_invert(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.xfail(reason="TODO")
def test_bitwise_or():
    """Test `bitwise_or`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=int), "")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=int), "")
    got = jnp.bitwise_or(x, y)
    expected = u.Quantity(jnp.bitwise_or(x.value, y.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.xfail(reason="TODO")
def test_bitwise_right_shift():
    """Test `bitwise_right_shift`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=int), "")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=int), "")
    got = jnp.bitwise_right_shift(x, y)
    expected = u.Quantity(jnp.bitwise_right_shift(x.value, y.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.xfail(reason="TODO")
def test_bitwise_xor():
    """Test `bitwise_xor`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=int), "")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=int), "")
    got = jnp.bitwise_xor(x, y)
    expected = u.Quantity(jnp.bitwise_xor(x.value, y.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_ceil():
    """Test `ceil`."""
    x = u.Quantity([1.1, 2.2, 3.3], "m")
    got = jnp.ceil(x)
    expected = u.Quantity(jnp.ceil(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_conj():
    """Test `conj`."""
    x = u.Quantity([1 + 2j, 3 + 4j], "m")
    got = jnp.conj(x)
    expected = u.Quantity(jnp.conj(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_cos():
    """Test `cos`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "rad")
    got = jnp.cos(x)
    expected = u.Quantity(jnp.cos(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_cosh():
    """Test `cosh`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "rad")
    got = jnp.cosh(x)
    expected = u.Quantity(jnp.cosh(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_divide():
    """Test `divide`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "km")
    got = jnp.divide(x, y)
    expected = u.Quantity(jnp.divide(x.value, y.value), "m / km")

    assert isinstance(got, u.Quantity)
    assert got.unit.is_equivalent(expected.unit)
    assert jnp.array_equal(got.value, expected.value)


def test_equal():
    """Test `equal`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.equal(x, y)
    expected = jnp.equal(x.value, y.value)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, expected)


def test_exp():
    """Test `exp`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "")
    got = jnp.exp(x)
    expected = u.Quantity(jnp.exp(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_expm1():
    """Test `expm1`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "")
    got = jnp.expm1(x)
    expected = u.Quantity(jnp.expm1(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_floor():
    """Test `floor`."""
    x = u.Quantity([1.1, 2.2, 3.3], "m")
    got = jnp.floor(x)
    expected = u.Quantity(jnp.floor(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_floor_divide():
    """Test `floor_divide`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.floor_divide(x, y)
    expected = u.Quantity(jnp.floor_divide(x.value, y.value), "m / m")

    assert isinstance(got, u.Quantity)
    assert got.unit.is_equivalent(expected.unit)
    assert jnp.array_equal(got.value, expected.value)


def test_greater():
    """Test `greater`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.greater(x, y)
    expected = jnp.greater(x.value, y.value)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, expected)


def test_greater_equal():
    """Test `greater_equal`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.greater_equal(x, y)
    expected = jnp.greater_equal(x.value, y.value)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, expected)


def test_imag():
    """Test `imag`."""
    x = u.Quantity([1 + 2j, 3 + 4j], "m")
    got = jnp.imag(x)
    expected = u.Quantity(jnp.imag(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_isfinite():
    """Test `isfinite`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.isfinite(x)
    expected = jnp.isfinite(x.value)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, expected)


def test_isinf():
    """Test `isinf`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.isinf(x)
    expected = jnp.isinf(x.value)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, expected)


def test_isnan():
    """Test `isnan`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.isnan(x)
    expected = jnp.isnan(x.value)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, expected)


def test_less():
    """Test `less`."""
    x = u.Quantity([1, 5, 3], "m")
    y = u.Quantity([4, 2, 6], "m")
    got = jnp.less(x, y)
    expected = jnp.less(x.value, y.value)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, expected)


def test_less_equal():
    """Test `less_equal`."""
    x = u.Quantity([1, 5, 3], "m")
    y = u.Quantity([4, 2, 6], "m")
    got = jnp.less_equal(x, y)
    expected = jnp.less_equal(x.value, y.value)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, expected)


def test_log():
    """Test `log`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "")
    got = jnp.log(x)
    expected = u.Quantity(jnp.log(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_log1p():
    """Test `log1p`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "")
    got = jnp.log1p(x)
    expected = u.Quantity(jnp.log1p(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_log2():
    """Test `log2`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "")
    got = jnp.log2(x)
    expected = u.Quantity(jnp.log2(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_log10():
    """Test `log10`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "")
    got = jnp.log10(x)
    expected = u.Quantity(jnp.log10(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.allclose(got, expected, atol=u.Quantity(1e-8, ""))


def test_logaddexp():
    """Test `logaddexp`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "")
    got = jnp.logaddexp(x, y)
    expected = u.Quantity(jnp.logaddexp(x.value, y.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_logical_and():
    """Test `logical_and`."""
    x = u.Quantity([True, False, True], "")
    y = u.Quantity([False, True, False], "")
    got = jnp.logical_and(x, y)
    expected = jnp.logical_and(x.value, y.value)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, expected)


def test_logical_not():
    """Test `logical_not`."""
    x = u.Quantity([True, False, True], "")
    got = jnp.logical_not(x)
    expected = u.Quantity(jnp.logical_not(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.xfail(reason="TODO")
def test_logical_or():
    """Test `logical_or`."""
    x = u.Quantity([True, False, True], "")
    y = u.Quantity([False, True, False], "")
    got = jnp.logical_or(x, y)
    expected = u.Quantity(jnp.logical_or(x.value, y.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.xfail(reason="TODO")
def test_logical_xor():
    """Test `logical_xor`."""
    x = u.Quantity([True, False, True], "")
    y = u.Quantity([False, True, False], "")
    got = jnp.logical_xor(x, y)
    expected = u.Quantity(jnp.logical_xor(x.value, y.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_multiply():
    """Test `multiply`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.multiply(x, y)
    expected = u.Quantity(jnp.multiply(x.value, y.value), "m2")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_negative():
    """Test `negative`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.negative(x)
    expected = u.Quantity(jnp.negative(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_not_equal():
    """Test `not_equal`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 2, 6], dtype=float), "m")
    got = jnp.not_equal(x, y)
    expected = jnp.not_equal(x.value, y.value)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, expected)


def test_positive():
    """Test `positive`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.positive(x)
    expected = u.Quantity(jnp.positive(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_pow_quantity_power():
    """Test `pow`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4], dtype=float), "")
    got = jnp.pow(x, y)
    expected = u.Quantity(jnp.pow(x.value, y.value), "m4")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_pow():
    """Test `pow`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = jnp.asarray([4], dtype=float)
    got = jnp.pow(x, y)
    expected = u.Quantity(jnp.pow(x.value, y), "m4")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_real():
    """Test `real`."""
    x = u.Quantity([1 + 2j, 3 + 4j], "m")
    got = jnp.real(x)
    expected = u.Quantity(jnp.real(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.skip(reason="TODO")
def test_remainder():
    """Test `remainder`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.remainder(x, y)
    expected = u.Quantity(jnp.remainder(x.value, y.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_round():
    """Test `round`."""
    x = u.Quantity([1.1, 2.2, 3.3], "m")
    got = jnp.round(x)
    expected = u.Quantity(jnp.round(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_sign():
    """Test `sign`."""
    x = u.Quantity([-1, 2, -3], "m")
    got = jnp.sign(x)
    expected = jnp.sign(x.value)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, expected)


def test_sin():
    """Test `sin`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "rad")
    got = jnp.sin(x)
    expected = u.Quantity(jnp.sin(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_sinh():
    """Test `sinh`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "rad")
    got = jnp.sinh(x)
    expected = u.Quantity(jnp.sinh(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_square():
    """Test `square`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.square(x)
    expected = u.Quantity(jnp.square(x.value), "m2")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_sqrt():
    """Test `sqrt`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.sqrt(x)
    expected = u.Quantity(jnp.sqrt(x.value), u.unit("m(1/2)"))

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_subtract():
    """Test `subtract`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.subtract(x, y)
    expected = u.Quantity(jnp.subtract(x.value, y.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_tan():
    """Test `tan`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "rad")
    got = jnp.tan(x)
    expected = u.Quantity(jnp.tan(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_tanh():
    """Test `tanh`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "rad")
    got = jnp.tanh(x)
    expected = u.Quantity(jnp.tanh(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.skip(reason="TODO")
def test_trunc():
    """Test `trunc`."""
    x = u.Quantity([1.1, 2.2, 3.3], "m")
    got = jnp.trunc(x)
    expected = u.Quantity(jnp.trunc(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# =============================================================================
# Indexing functions


@pytest.mark.skip(reason="TODO")
def test_take():
    """Test `take`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    indices = u.Quantity(jnp.asarray([0, 1, 2], dtype=int), "")
    got = jnp.take(x, indices)
    expected = u.Quantity(jnp.take(x.value, indices.value, axis=None), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# =============================================================================
# Linear algebra functions


@pytest.mark.skip(reason="TODO")
def test_matmul():
    """Test `matmul`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.matmul(x, y)
    expected = u.Quantity(jnp.matmul(x.value, y.value), "m2")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_matrix_transpose():
    """Test `matrix_transpose`."""
    x = u.Quantity(jnp.asarray([[1, 2, 3], [4, 5, 6]], dtype=float), "m")
    got = jnp.matrix_transpose(x)
    expected = u.Quantity(jnp.matrix_transpose(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.skip(reason="TODO")
def test_tensordot():
    """Test `tensordot`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    axes = 1
    got = jnp.tensordot(x, y, axes=axes)
    expected = u.Quantity(jnp.tensordot(x.value, y.value, axes=axes), "m2")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.skip(reason="TODO")
def test_vecdot():
    """Test `vecdot`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.vecdot(x, y)
    expected = u.Quantity(jnp.vecdot(x.value, y.value), "m2")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# =============================================================================
# Manipulation functions


def test_broadcast_arrays():
    """Test `broadcast_arrays`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4], dtype=float), "s")
    got = jnp.broadcast_arrays(x, y)
    expected = jnp.broadcast_arrays(x.value, y.value)

    assert isinstance(got, tuple | list)
    assert len(got) == len(expected)
    for got_, expected_ in zip(got, expected, strict=True):
        assert isinstance(got_, u.Quantity)
        assert jnp.array_equal(got_.value, expected_)


def test_broadcast_to():
    """Test `broadcast_to`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    shape = (2, 3)
    got = jnp.broadcast_to(x, shape)
    expected = u.Quantity(jnp.broadcast_to(x.value, shape), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_concat():
    """Test `concat`."""
    # TODO: test the axis argument
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4], dtype=float), "m")
    got = jnp.concat((x, y))
    expected = u.Quantity(jnp.concat((x.value, y.value)), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_expand_dims():
    """Test `expand_dims`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.expand_dims(x, axis=0)
    expected = u.Quantity(jnp.expand_dims(x.value, axis=0), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_flip():
    """Test `flip`."""
    x = u.Quantity(jnp.asarray([[1, 2, 3], [4, 5, 6]], dtype=float), "m")
    got = jnp.flip(x)
    expected = u.Quantity(jnp.flip(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_permute_dims():
    """Test `permute_dims`."""
    x = u.Quantity(jnp.asarray([[1, 2, 3], [4, 5, 6]], dtype=float), "m")
    got = jnp.permute_dims(x, (1, 0))
    expected = u.Quantity(jnp.permute_dims(x.value, (1, 0)), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_reshape():
    """Test `reshape`."""
    x = u.Quantity(jnp.asarray([[1, 2, 3], [4, 5, 6]], dtype=float), "m")
    got = jnp.reshape(x, (3, 2))
    expected = u.Quantity(jnp.reshape(x.value, (3, 2)), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_roll():
    """Test `roll`."""
    x = u.Quantity(jnp.asarray([[1, 2, 3], [4, 5, 6]], dtype=float), "m")
    got = jnp.roll(x, shift=1, axis=0)
    expected = u.Quantity(jnp.roll(x.value, shift=1, axis=0), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_squeeze():
    """Test `squeeze`."""
    x = u.Quantity(jnp.asarray([[[0], [1], [2]]], dtype=float), "m")
    got = jnp.squeeze(x, axis=(0, 2))
    expected = u.Quantity(jnp.squeeze(x.value, axis=(0, 2)), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_stack():
    """Test `stack`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    y = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.stack((x, y))
    expected = u.Quantity(jnp.stack((x.value, y.value)), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# =============================================================================
# Searching functions


def test_argmax():
    """Test `argmax`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.argmax(x)
    expected = u.Quantity(jnp.argmax(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_argmin():
    """Test `argmin`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.argmin(x)
    expected = u.Quantity(jnp.argmin(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.xfail(reason="TODO")
def test_nonzero():
    """Test `nonzero`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    (got,) = jnp.nonzero(x)
    (expected,) = u.Quantity(jnp.nonzero(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected)


def test_where():
    """Test `where`."""
    condition = u.Quantity(jnp.asarray([True, False, True]), "")
    y = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    z = u.Quantity(jnp.asarray([4, 5, 6], dtype=float), "m")
    got = jnp.where(condition, y, z)
    expected = u.Quantity(jnp.where(condition.value, y.value, z.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# =============================================================================
# Set functions


@pytest.mark.xfail(reason="value is not a Quantity")
def test_unique_all():
    """Test `unique_all`."""
    x = u.Quantity(jnp.asarray([1, 2, 2, 3, 3, 3], dtype=float))
    got = jnp.unique_all(x)
    expected = jnp.unique_all(x.value)

    assert isinstance(got, _UniqueAllResult)

    assert isinstance(got.values, u.Quantity)
    assert jnp.array_equal(got.values, expected.values)

    assert isinstance(got.inverse, u.Quantity)
    assert jnp.array_equal(got.inverse, expected.inverse)

    assert isinstance(got.inverse_indices, Array)
    assert jnp.array_equal(got.inverse_indices, expected.inverse_indices)

    assert isinstance(got.counts, Array)
    assert jnp.array_equal(got.counts, expected.counts)


@pytest.mark.xfail(reason="value is not a Quantity")
def test_unique_counts():
    """Test `unique_counts`."""
    x = u.Quantity(jnp.asarray([1, 2, 2, 3, 3, 3], dtype=float))
    got = jnp.unique_counts(x)
    expected = jnp.unique_counts(x.value)

    assert isinstance(got, _UniqueCountsResult)

    assert isinstance(got.values, u.Quantity)
    assert jnp.array_equal(got.values.value, expected.values)

    assert isinstance(got.counts, Array)
    assert jnp.array_equal(got.counts, expected.counts)


@pytest.mark.xfail(reason="value is not a Quantity")
def test_unique_inverse():
    """Test `unique_inverse`."""
    x = u.Quantity(jnp.asarray([1, 2, 2, 3, 3, 3], dtype=float))
    got = jnp.unique_inverse(x)
    expected = jnp.unique_inverse(x.value)

    assert isinstance(got, _UniqueInverseResult)

    assert isinstance(got.values, u.Quantity)
    assert jnp.array_equal(got.values.value, expected.values)

    assert isinstance(got.inverse, u.Quantity)
    assert jnp.array_equal(got.inverse.value, expected.inverse)


@pytest.mark.xfail(reason="value is not a Quantity")
def test_unique_values():
    """Test `unique_values`."""
    x = u.Quantity(jnp.asarray([1, 2, 2, 3, 3, 3], dtype=float))
    got = jnp.unique_values(x)
    expected = u.Quantity(jnp.unique_values(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# =============================================================================
# Sorting functions


def test_argsort():
    """Test `argsort`."""
    q = u.Quantity(jnp.asarray([3, 2, 1], dtype=float), "m")
    got = jnp.argsort(q)

    assert isinstance(got, Array)
    assert jnp.array_equal(got, jnp.argsort(q.value))


def test_sort():
    """Test `sort`."""
    q = u.Quantity(jnp.asarray([3, 2, 1], dtype=float), "m")
    got = jnp.sort(q)

    assert isinstance(got, u.Quantity)
    assert got.unit == u.unit("m")
    assert jnp.array_equal(got.value, jnp.sort(q.value))


# =============================================================================
# Statistical functions


def test_max():
    """Test `max`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.max(x)
    expected = u.Quantity(jnp.max(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.skip("TODO")
def test_mean():
    """Test `mean`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.mean(x)
    expected = u.Quantity(jnp.mean(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.skip("TODO")
def test_min():
    """Test `min`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.min(x)
    expected = u.Quantity(jnp.min(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.filterwarnings("ignore:Explicitly requested dtype")  # TODO: Why?
def test_prod():
    """Test `prod`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.prod(x)
    expected = u.Quantity(jnp.prod(x.value), "m3")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_std():
    """Test `std`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.std(x)
    expected = u.Quantity(jnp.std(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


@pytest.mark.filterwarnings("ignore:Explicitly requested dtype")  # TODO: Why?
def test_sum():
    """Test `sum`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.sum(x)
    expected = u.Quantity(jnp.sum(x.value), "m")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_var():
    """Test `var`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=float), "m")
    got = jnp.var(x)
    expected = u.Quantity(jnp.var(x.value), "m2")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# =============================================================================
# Utility functions


@pytest.mark.xfail(reason="returns a jax.Array")
def test_all():
    """Test `all`."""
    x = u.Quantity(jnp.asarray([True, False, True], dtype=bool), "")
    got = jnp.all(x)
    expected = u.Quantity(jnp.all(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


def test_any():
    """Test `any`."""
    x = u.Quantity(jnp.asarray([1, 2, 3], dtype=bool), "m")
    got = jnp.any(x)
    expected = u.Quantity(jnp.any(x.value), "")

    assert isinstance(got, u.Quantity)
    assert got.unit == expected.unit
    assert jnp.array_equal(got.value, expected.value)


# =============================================================================
# FFT


@pytest.mark.skip("TODO")
def test_fft_fft():
    """Test `fft.fft`."""


@pytest.mark.skip("TODO")
def test_fft_ifft():
    """Test `fft.ifft`."""


@pytest.mark.skip("TODO")
def test_fft_fftn():
    """Test `fft.fftn`."""


@pytest.mark.skip("TODO")
def test_fft_ifftn():
    """Test `fft.ifftn`."""


@pytest.mark.skip("TODO")
def test_fft_rfft():
    """Test `fft.rfft`."""


@pytest.mark.skip("TODO")
def test_fft_irfft():
    """Test `fft.irfft`."""


@pytest.mark.skip("TODO")
def test_fft_rfftn():
    """Test `fft.rfftn`."""


@pytest.mark.skip("TODO")
def test_fft_irfftn():
    """Test `fft.irfftn`."""


@pytest.mark.skip("TODO")
def test_fft_hfft():
    """Test `fft.hfft`."""


@pytest.mark.skip("TODO")
def test_fft_ihfft():
    """Test `fft.ihfft`."""


@pytest.mark.skip("TODO")
def test_fft_fftfreq():
    """Test `fft.fftfreq`."""


@pytest.mark.skip("TODO")
def test_fft_rfftfreq():
    """Test `fft.rfftfreq`."""


@pytest.mark.skip("TODO")
def test_fft_fftshift():
    """Test `fft.fftshift`."""


@pytest.mark.skip("TODO")
def test_fft_ifftshift():
    """Test `fft.ifftshift`."""


# =============================================================================
# Linalg


@pytest.mark.skip("TODO")
def test_linalg_cholesky():
    """Test `linalg.cholesky`."""


@pytest.mark.skip("TODO")
def test_linalg_cross():
    """Test `linalg.cross`."""


@pytest.mark.skip("TODO")
def test_linalg_det():
    """Test `linalg.det`."""


@pytest.mark.skip("TODO")
def test_linalg_diagonal():
    """Test `linalg.diagonal`."""


@pytest.mark.skip("TODO")
def test_linalg_eigh():
    """Test `linalg.eigh`."""


@pytest.mark.skip("TODO")
def test_linalg_eigvalsh():
    """Test `linalg.eigvalsh`."""


@pytest.mark.skip("TODO")
def test_linalg_inv():
    """Test `linalg.inv`."""


@pytest.mark.skip("TODO")
def test_linalg_matmul():
    """Test `linalg.matmul`."""


@pytest.mark.skip("TODO")
def test_linalg_matrix_norm():
    """Test `linalg.matrix_norm`."""


@pytest.mark.skip("TODO")
def test_linalg_matrix_power():
    """Test `linalg.matrix_power`."""


@pytest.mark.skip("TODO")
def test_linalg_matrix_rank():
    """Test `linalg.matrix_rank`."""


@pytest.mark.skip("TODO")
def test_linalg_matrix_transpose():
    """Test `linalg.matrix_transpose`."""


@pytest.mark.skip("TODO")
def test_linalg_outer():
    """Test `linalg.outer`."""


@pytest.mark.skip("TODO")
def test_linalg_pinv():
    """Test `linalg.pinv`."""


@pytest.mark.skip("TODO")
def test_linalg_qr():
    """Test `linalg.qr`."""


@pytest.mark.skip("TODO")
def test_linalg_slogdet():
    """Test `linalg.slogdet`."""


@pytest.mark.skip("TODO")
def test_linalg_solve():
    """Test `linalg.solve`."""


@pytest.mark.skip("TODO")
def test_linalg_svd():
    """Test `linalg.svd`."""


@pytest.mark.skip("TODO")
def test_linalg_svdvals():
    """Test `linalg.svdvals`."""


@pytest.mark.skip("TODO")
def test_linalg_tensordot():
    """Test `linalg.tensordot`."""


@pytest.mark.skip("TODO")
def test_linalg_trace():
    """Test `linalg.trace`."""


@pytest.mark.skip("TODO")
def test_linalg_vecdot():
    """Test `linalg.vecdot`."""


@pytest.mark.skip("TODO")
def test_linalg_vector_norm():
    """Test `linalg.vector_norm`."""


##############################################################################
# Functions


def test_allclose():
    """Test `e`."""
    q = u.Quantity(100.0, "m")

    with pytest.raises(Exception):  # noqa: B017, PT011
        assert jnp.allclose(q, u.Quantity(0.1, "km"))

    # Need the `atol` argument.
    assert jnp.allclose(q, u.Quantity(0.1, "km"), atol=u.Quantity(1e-6, "m"))



"""Orthogonal Mixin classes for Quantity classes."""

__all__: list[str] = []

from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, cast

import equinox as eqx
import numpy as np
from astropy.units import CompositeUnit
from jax.typing import ArrayLike
from jaxtyping import Array

from dataclassish import replace

from .api import uconvert, ustrip
from unxt._src.units import AstropyUnits
from unxt.units import unit as parse_unit

if TYPE_CHECKING:
    import unxt.quantity


class AstropyQuantityCompatMixin:
    """Mixin for compatibility with `astropy.units.Quantity`."""

    value: eqx.AbstractVar[ArrayLike]
    unit: eqx.AbstractVar[AstropyUnits]
    uconvert: Callable[[Any], "unxt.quantity.AbstractQuantity"]
    ustrip: Callable[[Any], ArrayLike]

    def to(self, u: Any, /) -> "unxt.quantity.AbstractQuantity":
        """Convert the quantity to the given units.

        See `unxt.quantity.AbstractQuantity.uconvert`.

        Examples
        --------
        >>> from unxt import Quantity

        >>> q = Quantity(1, "m")
        >>> q.to("cm")
        Quantity['length'](Array(100., dtype=float32, ...), unit='cm')

        """
        return uconvert(u, self)  # redirect to the standard method

    def to_value(self, u: Any, /) -> ArrayLike:
        """Return the value in the given units.

        See `unxt.AbstractQuantity.ustrip`.

        Examples
        --------
        >>> from unxt import Quantity

        >>> q = Quantity(1, "m")
        >>> q.to_value("cm")
        Array(100., dtype=float32, weak_type=True)

        """
        return ustrip(u, self)  # redirect to the standard method

    def decompose(
        self, bases: Sequence[AstropyUnits | str], /
    ) -> "unxt.quantity.AbstractQuantity":
        """Decompose the quantity into the given bases.

        Examples
        --------
        >>> from unxt import Quantity

        >>> q = Quantity(1, "m")
        >>> q.decompose(["cm", "s"])
        Quantity['length'](Array(100., dtype=float32, ...), unit='cm')

        """
        bases_ = [parse_unit(b) for b in bases]
        du = self.unit.decompose(bases_)  # decomposed units
        base_units = CompositeUnit(scale=1, bases=du.bases, powers=du.powers)
        return replace(self, value=self.value * du.scale, unit=base_units)


#####################################################################


SUPPORTED_IPYTHON_REPR_FORMATS: dict[str, str] = {
    "text/plain": "__repr__",
    "text/html": "_repr_html_",
    "text/latex": "_repr_latex_",
}


class IPythonReprMixin:
    """Mixin class for IPython representation of a Quantity."""

    value: Array
    unit: AstropyUnits

    def _repr_mimebundle_(
        self,
        *,
        include: Sequence[str] | None = None,
        exclude: Sequence[str] | None = None,
    ) -> dict[str, str]:
        r"""Return a MIME bundle representation of the Quantity.

        :param include: The set of keys to include in the MIME bundle. If not
            provided, all supported formats are included.
        :param exclude: The set of keys to exclude in the MIME bundle. If not
            provided, all supported formats are included. 'include' has
            precedence over 'exclude'.

        Examples
        --------
        >>> from unxt import Quantity

        >>> q = Quantity([1.0, 2, 3, 4], "m")
        >>> q._repr_mimebundle_()
        {'text/plain': "Quantity['length'](Array([1., 2., 3., 4.], dtype=float32), unit='m')",
         'text/html': '<span>[1., 2., 3., 4.]</span> * <span>Unit("m")</span>',
         'text/latex': '$[1.,~2.,~3.,~4.] \\; \\mathrm{m}$'}

        >>> q._repr_mimebundle_(include=["text/plain"])
        {'text/plain': "Quantity['length'](Array([1., 2., 3., 4.], dtype=float32), unit='m')"}

        >>> q._repr_mimebundle_(exclude=["text/html", "text/latex"])
        {'text/plain': "Quantity['length'](Array([1., 2., 3., 4.], dtype=float32), unit='m')"}

        """  # noqa: E501
        # Determine the set of keys to include in the MIME bundle
        keys: Sequence[str]
        if include is None and exclude is None:
            keys = tuple(SUPPORTED_IPYTHON_REPR_FORMATS)
        elif include is not None:
            keys = [key for key in include if key in SUPPORTED_IPYTHON_REPR_FORMATS]
        else:
            keys = [
                k for k in SUPPORTED_IPYTHON_REPR_FORMATS if k not in cast(str, exclude)
            ]

        # Create the MIME bundle
        return {
            key: getattr(self, SUPPORTED_IPYTHON_REPR_FORMATS[key])() for key in keys
        }

    def _repr_html_(self) -> str:
        """Return an HTML representation of the Quantity.

        Examples
        --------
        >>> from unxt import Quantity

        >>> q = Quantity([1.0, 2, 3, 4], "m")
        >>> q._repr_html_()
        '<span>[1., 2., 3., 4.]</span> * <span>Unit("m")</span>'

        """
        unit_repr = getattr(self.unit, "_repr_html_", self.unit.__repr__)()
        value_repr = np.array2string(self.value, separator=", ")  # type: ignore[call-overload]

        return f"<span>{value_repr}</span> * <span>{unit_repr}</span>"

    def _repr_latex_(self) -> str:
        r"""Return a LaTeX representation of the Quantity.

        Examples
        --------
        >>> from unxt import Quantity

        >>> q = Quantity([1.0, 2, 3, 4], "m")
        >>> q._repr_latex_()
        '$[1.,~2.,~3.,~4.] \\; \\mathrm{m}$'

        """
        unit_repr = getattr(self.unit, "_repr_latex_", self.unit.__repr__)()
        value_repr = np.array2string(self.value, separator=",~")  # type: ignore[call-overload]

        return f"${value_repr} \\; {unit_repr[1:-1]}$"

    # TODO: implement:
    # - _repr_markdown_
    # - _repr_json_


#####################################################################


class NumPyCompatMixin:
    """Mixin for compatibility with numpy arrays."""

    unit: AstropyUnits

    __array_namespace__: Callable[[], Any]

    def __array__(self, **kwargs: Any) -> np.typing.NDArray[Any]:
        """Return the array as a numpy array, stripping the units.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import numpy as np

        >>> q = Quantity(1.01, "m")
        >>> np.array(q)
        array(1.01, dtype=float32)

        """
        return np.asarray(ustrip(self.unit, self), **kwargs)

    # TODO: why doesn't `__array_namespace__` supersede this?
    def __array_function__(
        self,
        func: Callable[..., Any],
        types: Any,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Any:
        """Dispatch to the corresponding jax.numpy function.

        Examples
        --------
        >>> import numpy as np
        >>> import unxt as u

        >>> q = u.Quantity([1.0, 2, 3, 4], "m")
        >>> np.sum(q)
        Quantity['length'](Array(10., dtype=float32), unit='m')

        >>> np.stack([q, q])
        Quantity['length'](Array([[1., 2., 3., 4.],
                                  [1., 2., 3., 4.]], dtype=float32), unit='m')

        """
        xp = self.__array_namespace__()
        xfunc = getattr(xp, func.__name__)
        return xfunc(*args, **kwargs)



"""Comparison functions."""

__all__ = ["equivalent"]

from plum import dispatch

from .base import AbstractUnitSystem


@dispatch
def equivalent(a: AbstractUnitSystem, b: AbstractUnitSystem, /) -> bool:
    """Check if two unit systems are equivalent."""
    return set(a.base_dimensions) == set(b.base_dimensions)



"""Units objects in JAX.

Copyright (c) 2023 Galactic Dynamics. All rights reserved.
"""

__all__ = ["unit", "unit_of"]

from typing import Any, TypeAlias

import astropy.units as apyu
from plum import dispatch

AstropyUnits: TypeAlias = apyu.Unit | apyu.UnitBase | apyu.CompositeUnit


@dispatch.abstract
def unit(obj: Any, /) -> AstropyUnits:
    """Construct the units from a units object."""


@dispatch.abstract
def unit_of(obj: Any, /) -> AstropyUnits:
    """Return the units of an object."""



"""Unitsystem compatibility."""

__all__: list[str] = []


import astropy.units as apyu
from plum import dispatch

from unxt.dims import AbstractDimension


@dispatch
def dimension_of(obj: apyu.Quantity, /) -> AbstractDimension:
    """Return the dimension of a quantity.

    Examples
    --------
    >>> import unxt as u
    >>> import astropy.units as apyu

    >>> q = apyu.Quantity(1, "m")
    >>> u.dimension_of(q)
    PhysicalType('length')

    """
    return dimension_of(obj.unit)



"""Test the `unxt.unitsystems` module."""

import itertools
import pickle
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated

import astropy.units as apyu
import numpy as np
import pytest

from unxt import dimension, unit, unitsystems
from unxt._src.unitsystems.base import _UNITSYSTEMS_REGISTRY
from unxt.unitsystems import (
    AbstractUnitSystem,
    AbstractUSysFlag,
    DimensionlessUnitSystem,
    DynamicalSimUSysFlag,
    StandardUSysFlag,
    dimensionless,
    equivalent,
    unitsystem,
)


@pytest.fixture
def clean_unitsystems_registry(monkeypatch):
    clean_registry = {}
    monkeypatch.setattr(
        "unxt._src.unitsystems.base._UNITSYSTEMS_REGISTRY", clean_registry
    )
    return clean_registry


# ===================================================================


def test_unitsystem_from_() -> None:
    """Test the `unxt.AbstractUnitSystem.from_`."""
    usys = unitsystem(5 * apyu.kpc, 50 * apyu.Myr, 1e5 * apyu.Msun, "rad")
    assert np.isclose((8 * apyu.Myr).decompose(usys).value, 8 / 50)


def test_compare() -> None:
    """Test the `unxt.AbstractUnitSystem.compare` method."""
    usys1 = unitsystem("kpc", "Myr", "radian", "Msun", "mas / yr")
    usys1_clone = unitsystem("kpc", "Myr", "radian", "Msun", "mas / yr")

    usys2 = unitsystem("kpc", "Myr", "radian", "Msun", "kiloarcsecond / yr")
    usys3 = unitsystem("kpc", "Myr", "radian", "kg", "mas / yr")

    assert usys1 == usys1_clone
    assert usys1_clone == usys1

    assert usys1 != usys2
    assert usys2 != usys1

    assert usys1 != usys3
    assert usys3 != usys1


def test_regression_dimension_aliases_spaces() -> None:
    usys = unitsystem("kpc", "Myr", "radian", "Msun", "mas / yr")
    assert usys["angular speed"] == usys["angular velocity"]


def test_pickle(tmpdir: Path) -> None:
    """Test pickling and unpickling a `unxt.AbstractUnitSystem`."""
    usys = unitsystem("kpc", "Myr", "radian", "Msun")

    path = tmpdir / "test.pkl"
    with path.open(mode="wb") as f:
        pickle.dump(usys, f)

    with path.open(mode="rb") as f:
        usys2 = pickle.load(f)  # noqa: S301

    assert usys == usys2


def test_non_frozen(clean_unitsystems_registry):
    # Passes
    class NoFrozen1(AbstractUnitSystem):
        pass

    clean_unitsystems_registry.clear()

    # Fails
    with pytest.raises(TypeError, match="cannot inherit non-frozen"):

        @dataclass(slots=False)
        class NoFrozen2(AbstractUnitSystem):
            pass

    clean_unitsystems_registry.clear()

    # Passes
    @dataclass(frozen=True, slots=True)
    class NoFrozen3(AbstractUnitSystem):
        pass


@pytest.mark.usefixtures("clean_unitsystems_registry")
def test_non_unit_fields():
    """Test that non-Unit fields are skipped."""

    @dataclass(frozen=True, slots=True)
    class SomeNoneUnitFields(AbstractUnitSystem):
        a: Annotated[apyu.Unit, dimension("length")]
        b: int

    assert SomeNoneUnitFields._base_field_names == ("a",)


@pytest.mark.usefixtures("clean_unitsystems_registry")
def test_wrong_annotation():
    """Test that non-Unit fields are skipped."""
    # No dimension annotation
    with pytest.raises(
        TypeError, match="Field 'a' must be an Annotated with a dimension"
    ):

        @dataclass(frozen=True, slots=True)
        class BadAnnotations(AbstractUnitSystem):
            a: Annotated[apyu.Unit, "no dimension annotation"]

    # Too many dimension annotations
    match = "Field 'a' must be an Annotated with only one dimension"
    with pytest.raises(TypeError, match=match):

        @dataclass(frozen=True, slots=True)
        class BadAnnotations(AbstractUnitSystem):
            a: Annotated[apyu.Unit, dimension("length"), dimension("time")]


def test_unitsystem_already_registered():
    """Test that a unit system can only be registered once."""

    class MyUnitSystem(AbstractUnitSystem):
        length: Annotated[apyu.Unit, dimension("length")]
        time: Annotated[apyu.Unit, dimension("time")]

    assert MyUnitSystem._base_dimensions in unitsystems.UNITSYSTEMS_REGISTRY

    with pytest.raises(ValueError, match="already exists"):

        class MyUnitSystem(AbstractUnitSystem):
            length: Annotated[apyu.Unit, dimension("length")]
            time: Annotated[apyu.Unit, dimension("time")]

    # Clean up custom unit system from registry:
    del _UNITSYSTEMS_REGISTRY[MyUnitSystem._base_dimensions]


class TestDimensionlessUnitSystem:
    """Test `unxt.unitsystems.DimensionlessUnitSystem`."""

    def test_getitem(self) -> None:
        """Test :meth:`unxt.unitsystems.DimensionlessUnitSystem.__getitem__`."""
        assert dimensionless["dimensionless"] == apyu.one

        with pytest.raises(apyu.UnitConversionError):
            dimensionless["length"]

    def test_decompose(self) -> None:
        """Test that dimensionless unitsystem can be decomposed."""
        with pytest.raises(ValueError, match="can not be decomposed into"):
            (15 * apyu.kpc).decompose(dimensionless)

    def test_str(self) -> None:
        """Test that the string representation is correct."""
        assert str(dimensionless) == "DimensionlessUnitSystem()"


def test_dimensionless_singleton():
    """Test that the dimensionless unit system is a singleton."""
    assert DimensionlessUnitSystem() is dimensionless


def test_equivalent():
    """Test that equivalent unit systems are equal."""
    usys1 = unitsystem("kpc", "Myr", "radian", "Msun", "mas / yr")
    usys2 = unitsystem("km", "yr", "deg", "kg", "deg / s")
    assert equivalent(usys1, usys2)

    usys3 = unitsystem("kpc", "Myr", "radian")
    assert not equivalent(usys1, usys3)


def test_extend():
    """Test adding additional units to a unit system."""
    usys1 = unitsystem("kpc", "Myr", "radian", "Msun", "km / s")
    usys2 = unitsystem(usys1, "mas / yr")
    assert usys2["angular speed"] == unit("mas / yr")

    usys3 = unitsystem(usys1, "mas / yr", "pc")
    assert usys3["angular speed"] == unit("mas / yr")
    assert usys3["length"] == unit("pc")  # overridden


def test_abstract_usys_flag():
    """Test that the abstract unit system flag fails."""
    with pytest.raises(TypeError, match="Do not use"):
        unitsystem(AbstractUSysFlag, "kpc")

    with pytest.raises(ValueError, match="unit system flag classes"):
        AbstractUSysFlag()


def test_standard_flag():
    """Test defining unit system with the standard flag."""
    usys1 = unitsystem(StandardUSysFlag, "kpc", "Myr")
    usys2 = unitsystem("kpc", "Myr")
    assert usys1 == usys2

    with pytest.raises(ValueError, match="unit system flag classes"):
        StandardUSysFlag()


def test_simulation_usys():
    """Test defining the simulation unit system with expected inputs."""
    from astropy.constants import G as const_G  # noqa: N811

    tmp_G = const_G.decompose([apyu.kpc, apyu.Myr, apyu.Msun])
    usys1 = unitsystem(DynamicalSimUSysFlag, "kpc", "Myr", "rad")
    assert np.isclose((1 * usys1["mass"]).to_value("Msun"), 1 / tmp_G.value)

    usys2 = unitsystem(DynamicalSimUSysFlag, "kpc", "Msun", "rad")
    assert np.isclose((1 * usys2["time"]).to_value("Myr"), 1 / np.sqrt(tmp_G.value))

    base_units = ("kpc", "Myr", "Msun", "km / s")
    for u1, u2 in itertools.product(base_units, base_units):
        if u1 == u2:
            continue

        usys = unitsystem(DynamicalSimUSysFlag, u1, u2)

        # For now, just test retrieving all three base unit types:
        usys["length"]
        usys["mass"]
        usys["time"]



"""Quantities in JAX."""

__all__ = [
    "AbstractParametricQuantity",
    "AbstractQuantity",
    "Quantity",
    "BareQuantity",
    "UncheckedQuantity",
    "is_unit_convertible",
    "uconvert",
    "ustrip",
    "is_any_quantity",
    "convert_to_quantity_value",
]

from .api import is_unit_convertible, uconvert, ustrip
from .base import AbstractQuantity, is_any_quantity
from .base_parametric import AbstractParametricQuantity
from .quantity import Quantity
from .unchecked import BareQuantity, UncheckedQuantity
from .value import convert_to_quantity_value



"""Built-in unit systems."""

from __future__ import annotations

__all__ = ["DimensionlessUnitSystem", "LTMAUnitSystem"]

from dataclasses import dataclass
from typing import Annotated, TypeAlias, final
from typing_extensions import override

from astropy.units import UnitBase as AstropyUnitBase, dimensionless_unscaled

from . import builtin_dimensions as ud
from .base import AbstractUnitSystem
from unxt._src.utils import SingletonMixin

Unit: TypeAlias = AstropyUnitBase


@final
@dataclass(frozen=True, slots=True)
class DimensionlessUnitSystem(SingletonMixin, AbstractUnitSystem):
    """A unit system with only dimensionless units.

    This is a singleton class.

    Examples
    --------
    >>> from unxt.unitsystems import DimensionlessUnitSystem
    >>> dims1 = DimensionlessUnitSystem()
    >>> dims2 = DimensionlessUnitSystem()
    >>> dims1 is dims2
    True

    """

    #: The dimensionless unit.
    dimensionless: Annotated[Unit, ud.dimensionless] = dimensionless_unscaled

    def __repr__(self) -> str:
        return "DimensionlessUnitSystem()"

    @override
    def __str__(self) -> str:
        return self.__repr__()


@final
@dataclass(frozen=True, slots=True, repr=False)
class LTMAUnitSystem(AbstractUnitSystem):
    """Length, time, mass, angle unit system."""

    #: Units for the length dimension.
    length: Annotated[Unit, ud.length]

    #: Units for the time dimension.
    time: Annotated[Unit, ud.time]

    #: Units for the mass dimension.
    mass: Annotated[Unit, ud.mass]

    #: Units for the angle 'dimension'.
    angle: Annotated[Unit, ud.angle]

    def __repr__(self) -> str:
        fs = ", ".join(map(str, self.base_units))
        return f"unitsystem({fs})"


@final
@dataclass(frozen=True, slots=True)
class SIUnitSystem(SingletonMixin, AbstractUnitSystem):
    """SI unit system + angles.

    Note: this is not part of the public API! Use the `si` instance (realization) from
    `unxt.unitsystems` instead.

    Examples
    --------
    >>> import unxt as u
    >>> from unxt._src.unitsystems.builtin import SIUnitSystem
    >>> usys = SIUnitSystem(
    ...     length=u.unit("meter"),
    ...     time=u.unit("second"),
    ...     mass=u.unit("kilogram"),
    ...     electric_current=u.unit("ampere"),
    ...     temperature=u.unit("Kelvin"),
    ...     amount=u.unit("mole"),
    ...     luminous_intensity=u.unit("candela"),
    ...     angle=u.unit("radian"),
    ... )
    >>> usys
    unitsystem(m, kg, s, mol, A, K, cd, rad)

    """

    # Base SI dimensions
    #: Units for the length dimension.
    length: Annotated[Unit, ud.length]

    #: Units for the mass dimension.
    mass: Annotated[Unit, ud.mass]

    #: Units for the time dimension.
    time: Annotated[Unit, ud.time]

    #: Units for the amount of substance dimension.
    amount: Annotated[Unit, ud.amount]

    #: Units for the electric current dimension.
    electric_current: Annotated[Unit, ud.current]

    #: Units for the temperature dimension.
    temperature: Annotated[Unit, ud.temperature]

    #: Units for the luminous intensity dimension.
    luminous_intensity: Annotated[Unit, ud.luminous_intensity]

    # + angles
    #: Units for the angle 'dimension'.
    angle: Annotated[Unit, ud.angle]

    def __repr__(self) -> str:
        fs = ", ".join(map(str, self.base_units))
        return f"unitsystem({fs})"


@final
@dataclass(frozen=True, slots=True)
class CGSUnitSystem(SingletonMixin, AbstractUnitSystem):
    """CGS unit system + angles.

    Note: this is not part of the public API! Use the `cgs` instance (realization) from
    `unxt.unitsystems` instead.

    Examples
    --------
    >>> import unxt as u
    >>> from unxt._src.unitsystems.builtin import CGSUnitSystem
    >>> usys = CGSUnitSystem(
    ...     length=u.unit("centimeter"),
    ...     time=u.unit("second"),
    ...     mass=u.unit("gram"),
    ...     angle=u.unit("radian"),
    ...     force=u.unit("dyne"),
    ...     energy=u.unit("erg"),
    ...     pressure=u.unit("barye"),
    ...     dynamic_viscosity=u.unit("poise"),
    ...     kinematic_viscosity=u.unit("stokes"),
    ... )
    >>> usys
    unitsystem(cm, g, s, dyn, erg, Ba, P, St, rad)

    """

    # Base CGS dimensions
    #: Units for the length dimension.
    length: Annotated[Unit, ud.length]

    #: Units for the mass dimension.
    mass: Annotated[Unit, ud.mass]

    #: Units for the time dimension.
    time: Annotated[Unit, ud.time]

    #: Units for the force dimension.
    force: Annotated[Unit, ud.force]

    #: Units for the energy dimension.
    energy: Annotated[Unit, ud.energy]

    #: Units for the pressure dimension.
    pressure: Annotated[Unit, ud.pressure]

    #: Units for the dynamic viscosity dimension.
    dynamic_viscosity: Annotated[Unit, ud.dynamic_viscosity]

    #: Units for the kinematic viscosity dimension.
    kinematic_viscosity: Annotated[Unit, ud.kinematic_viscosity]

    # + angles
    #: Units for the angle 'dimension'.
    angle: Annotated[Unit, ud.angle]

    def __repr__(self) -> str:
        fs = ", ".join(map(str, self.base_units))
        return f"unitsystem({fs})"
